@@ -1,9 +1,12 @@
-from aisquared.base import BaseObject
-from aisquared.config.harvesting import ImageHarvester, TextHarvester
-from aisquared.config.preprocessing import TabularPreprocessor, ImagePreprocessor, TextPreprocessor
+from aisquared.base import BaseObject, ALLOWED_STAGES
+from aisquared.config.harvesting import ImageHarvester, TextHarvester, InputHarvester
+from aisquared.config.preprocessing.tabular import TabularPreprocessor
+from aisquared.config.preprocessing.image import ImagePreprocessor
+from aisquared.config.preprocessing.text import TextPreprocessor
 from aisquared.config.analytic import DeployedAnalytic, DeployedModel, LocalModel, LocalAnalytic
 from aisquared.config.postprocessing import BinaryClassification, MulticlassClassification, ObjectDetection, Regression
-from aisquared.config.rendering import ImageRendering, ObjectRendering, DocumentRendering, WordRendering
+from aisquared.config.rendering import ImageRendering, ObjectRendering, DocumentRendering, WordRendering, FilterRendering
+from aisquared.config.feedback import SimpleFeedback, BinaryFeedback, MulticlassFeedback, RegressionFeedback, ModelFeedback, QualitativeFeedback
 
 import tensorflowjs as tfjs
 import tensorflow as tf
@@ -13,7 +16,8 @@ import os
 
 HARVESTING_CLASSES = (
     ImageHarvester,
-    TextHarvester
+    TextHarvester,
+    InputHarvester
 )
 
 PREPROCESSING_CLASSES = (
@@ -40,7 +44,17 @@ RENDERING_CLASSES = (
     ObjectRendering,
     ImageRendering,
     DocumentRendering,
-    WordRendering
+    WordRendering,
+    FilterRendering
+)
+
+FEEDBACK_CLASSES = (
+    ModelFeedback,
+    SimpleFeedback,
+    BinaryFeedback,
+    MulticlassFeedback,
+    RegressionFeedback,
+    QualitativeFeedback
 )
 
 LOCAL_CLASSES = (
@@ -60,11 +74,16 @@ class ModelConfiguration(BaseObject):
             analytic,
             postprocessing_steps,
             rendering_steps,
+            feedback_steps = None,
+            stage = ALLOWED_STAGES[0],
             version = None,
             description = '',
             mlflow_uri = None,
             mlflow_user = None,
-            mlflow_token = None            
+            mlflow_token = None,
+            owner = None,
+            url = '*',
+            auto_run = False
     ):
         """
         Parameters
@@ -79,8 +98,12 @@ class ModelConfiguration(BaseObject):
             Analytics to use
         postprocessing_steps : Postprocessing object or list of Postprocessing objects or None
             Postprocessers to use
-        rendering_steps : Rendering object or list of Rendering objects
+        rendering_steps : Rendering object or list of Rendering objects or None
             Renderers to use
+        feedback_steps : None, Feedback object or list of Feedback objects or None (default None)
+            Feedback steps to use
+        stage : str (default 'experimental')
+            The stage of the model, from 'experimental', 'staging', 'production'
         version : str or None (default None)
             Version of the analytic
         description : str (default '')
@@ -91,6 +114,12 @@ class ModelConfiguration(BaseObject):
             MLFlow user to use, if applicable
         mlflow_token : str or None (default None)
             MLFlow token to use, if applicable
+        owner : str or None (default None)
+            The owner of the model
+        url : str (default '*')
+            URL or URL pattern to match
+        auto_run : bool (default False)
+            Whether to automatically run this file when on a valid page
         """
         super().__init__()
         self.name = name
@@ -99,11 +128,16 @@ class ModelConfiguration(BaseObject):
         self.analytic = analytic
         self.postprocessing_steps = postprocessing_steps
         self.rendering_steps = rendering_steps
+        self.stage = stage
+        self.feedback_steps = feedback_steps
         self.version = version
         self.description = description
         self.mlflow_uri = mlflow_uri
         self.mlflow_user = mlflow_user
         self.mlflow_token = mlflow_token
+        self.owner = owner
+        self.url = url
+        self.auto_run = auto_run
 
     # name
     @property
@@ -119,13 +153,14 @@ class ModelConfiguration(BaseObject):
         return self._harvesting_steps
     @harvesting_steps.setter
     def harvesting_steps(self, value):
+        harvesting_classes = HARVESTING_CLASSES + (ModelConfiguration,)
         if value is None:
             self._harvesting_steps = value
-        elif isinstance(value, HARVESTING_CLASSES):
+        elif isinstance(value, harvesting_classes):
             self._harvesting_steps = [value]
-        elif isinstance(value, list) and all([isinstance(val, HARVESTING_CLASSES) for val in value]):
+        elif isinstance(value, list) and all([isinstance(val, harvesting_classes) for val in value]):
             self._harvesting_steps = value
-        elif isinstance(value, list) and all([isinstance(val, list) for val in value]) and all([isinstance(v, HARVESTING_CLASSES) for val in value for v in val]):
+        elif isinstance(value, list) and all([isinstance(val, list) for val in value]) and all([isinstance(v, harvesting_classes) for val in value for v in val]):
             self._harvesting_steps = value
         else:
             raise ValueError('harvesting_steps must be a None, single Harvester object, a list of Harvester objects, or a list of list of harvester objects')
@@ -183,23 +218,48 @@ class ModelConfiguration(BaseObject):
         else:
             raise ValueError('postprocessing_steps must be a single Postprocessing object, a list of Postprocessing objects, or a list of list of Postprocessing objects')
 
-
     # rendering_steps
     @property
     def rendering_steps(self):
         return self._rendering_steps
     @rendering_steps.setter
     def rendering_steps(self, value):
-        if isinstance(value, RENDERING_CLASSES):
+        if isinstance(value, RENDERING_CLASSES) or value is None:
             self._rendering_steps = [value]
         elif isinstance(value, list) and all([isinstance(val, RENDERING_CLASSES) for val in value]):
             self._rendering_steps = value
         elif isinstance(value, list) and all([isinstance(val, list) for val in value]) and all([isinstance(v, RENDERING_CLASSES) for val in value for v in val]):
             self._rendering_steps = value
         else:
-            raise ValueError('rendering_steps must be a single Rendering object, a list of Rendering objects, or a list of list of Rendering objects')
+            raise ValueError('rendering_steps must be a single Rendering object, a list of Rendering objects, a list of list of Rendering objects, or None')
 
+    # feedback_steps
+    @property
+    def feedback_steps(self):
+        return self._feedback_steps
+    @feedback_steps.setter
+    def feedback_steps(self, value):
+        if value is None:
+            self._feedback_steps = value
+        elif isinstance(value, FEEDBACK_CLASSES):
+            self._feedback_steps = [value]
+        elif isinstance(value, list) and all([isinstance(val, FEEDBACK_CLASSES) for val in value]):
+            self._feedback_steps = value
+        elif isinstance(value, list) and all([isinstance(val, list) for val in value]) and all([isinstance(v, FEEDBACK_CLASSES) for val in value for v in val]):
+            self._feedback_steps = value
+        else:
+            raise ValueError('feedback_steps must be a single Feedback object, a list of Feedback objects, a list of list of Feedback objects, or None')
 
+    # stage
+    @property
+    def stage(self):
+        return self._stage
+    @stage.setter
+    def stage(self, value):
+        if value not in ALLOWED_STAGES:
+            raise ValueError(f'stage must be one of {ALLOWED_STAGES}')
+        self._stage = value
+        
     # version
     @property
     def version(self):
@@ -240,12 +300,43 @@ class ModelConfiguration(BaseObject):
     def mlflow_token(self, value):
         self._mlflow_token = value
 
+    # owner
+    @property
+    def owner(self):
+        return self._owner
+    @owner.setter
+    def owner(self, value):
+        if not isinstance(value, str) and value is not None:
+            raise ValueError('owner must be a string or None')
+        self._owner = value
+
+    # url
+    @property
+    def url(self):
+        return self._url
+    @url.setter
+    def url(self, value):
+        if not isinstance(value, str):
+            raise ValueError('url must be string')
+        self._url = value
+    
+    # auto_run
+    @property
+    def auto_run(self):
+        return self._auto_run
+    @auto_run.setter
+    def auto_run(self, value):
+        if not isinstance(value, bool):
+            raise TypeError('auto_run must be Boolean valued')
+        self._auto_run = value
+
     # harvester_dict
     @property
     def harvester_dict(self):
+        harvesting_classes = HARVESTING_CLASSES + (ModelConfiguration,)
         if self.harvesting_steps is None:
             return None
-        elif isinstance(self.harvesting_steps, list) and all([isinstance(val, HARVESTING_CLASSES) for val in self.harvesting_steps]):
+        elif isinstance(self.harvesting_steps, list) and all([isinstance(val, harvesting_classes) for val in self.harvesting_steps]):
             return [val.to_dict() for val in self.harvesting_steps]
         else:
             return [
@@ -289,18 +380,40 @@ class ModelConfiguration(BaseObject):
     # render_dict
     @property
     def render_dict(self):
-        if isinstance(self.rendering_steps, list) and all([isinstance(val, RENDERING_CLASSES) for val in self.rendering_steps]):
+        if self.rendering_steps[0] is None:
+            return []
+        elif isinstance(self.rendering_steps, list) and all([isinstance(val, RENDERING_CLASSES) for val in self.rendering_steps]):
             return [val.to_dict() for val in self.rendering_steps]
         else:
             return [
                 [v.to_dict() for v in val] for val in self.rendering_steps
             ]
 
+    # feedback_dict
+    @property
+    def feedback_dict(self):
+        if self.feedback_steps is None:
+            return self.feedback_steps
+        elif isinstance(self.feedback_steps, list) and all([isinstance(val, FEEDBACK_CLASSES) for val in self.feedback_steps]):
+            return [val.to_dict() for val in self.feedback_steps]
+        else:
+            return [
+                [v.to_dict() for v in val] for val in self.feedback_steps
+            ]
+
     def get_model_filenames(self):
         """
         Get filenames for all models in the configuration
         """
         filenames = []
+        if isinstance(self.harvesting_steps[0], list):
+            harvesting_list = [h for harvester in self.harvesting_steps for h in harvester]
+        else:
+            harvesting_list = self.harvesting_steps
+        for harvester in harvesting_list:
+            if isinstance(harvester, ModelConfiguration):
+                filenames.extend(harvester.get_model_filenames())
+
         if isinstance(self.analytic[0], ANALYTIC_CLASSES):
             for a in self.analytic:
                 if isinstance(a, LOCAL_CLASSES):
@@ -325,15 +438,20 @@ class ModelConfiguration(BaseObject):
                 'analytics' : self.analytic_dict,
                 'postprocessingSteps' : self.postprocesser_dict,
                 'renderingSteps' : self.render_dict,
+                'feedbackSteps' : self.feedback_dict,
+                'stage' : self.stage,
                 'version' : self.version,
                 'description' : self.description,
                 'mlflowUri' : self.mlflow_uri,
                 'mlflowUser' : self.mlflow_user,
-                'mlflowToken' : self.mlflow_token
+                'mlflowToken' : self.mlflow_token,
+                'owner' : self.owner,
+                'url' : self.url,
+                'autoRun' : self.auto_run
             }
         }
 
-    def compile(self, filename = None):
+    def compile(self, filename = None, dtype = None):
         """
         Compile the object into a '.air' file, which can then be dragged and 
         dropped into applications using the AI Squared JavaScript SDK
@@ -343,10 +461,17 @@ class ModelConfiguration(BaseObject):
         filename : path-like or None (default None)
             Filename to compile to. If None, defaults to '{NAME}.air', where {NAME} is the 
             name of the analytic
+        dtype : str or None (default None)
+            The datatype to use for the model weights. If None, defaults to 'float32'
         """
         if filename is None:
             filename = self.name + '.air'
 
+        if dtype is None:
+            dtype_map = None
+        else:
+            dtype_map = {dtype : '*'}
+            
         dirname = os.path.join('.', os.path.splitext(filename)[0])
 
         # write the object as json config 
@@ -361,7 +486,7 @@ class ModelConfiguration(BaseObject):
                 if os.path.splitext(f)[-1] == '.h5':
                     model = tf.keras.models.load_model(f)
                     model_dir = os.path.join(dirname, os.path.split(f)[-1])
-                    tfjs.converters.save_keras_model(model, model_dir)
+                    tfjs.converters.save_keras_model(model, model_dir, quantization_dtype_map = dtype_map)
                 else:
                     shutil.copy(f, dirname)
         
